# Generated by Django 3.2.22 on 2024-01-29 13:24

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('integrated_channel', '0031_alter_integratedchannelapirequestlogs_options'),
    ]

    operations = [
        migrations.AlterField(
            model_name='integratedchannelapirequestlogs',
            name='endpoint',
            field=models.URLField(max_length=255),
        ),
    ]
