# Generated by Django 4.2.15 on 2024-09-02 08:39

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('canvas', '0037_canvasenterprisecustomerconfiguration_copy_id_and_secret_and_more'),
    ]

    operations = [
        migrations.AlterField(
            model_name='canvasenterprisecustomerconfiguration',
            name='client_id',
            field=models.CharField(blank=True, default='', help_text='The API Client ID provided to edX by the enterprise customer to be used to make API calls to Canvas on behalf of the customer.', max_length=255, null=True, verbose_name='API Client ID'),
        ),
        migrations.AlterField(
            model_name='canvasenterprisecustomerconfiguration',
            name='client_secret',
            field=models.CharField(blank=True, default='', help_text='The API Client Secret provided to edX by the enterprise customer to be used to make  API calls to Canvas on behalf of the customer.', max_length=255, null=True, verbose_name='API Client Secret'),
        ),
    ]
