# Generated by Django 3.2.20 on 2023-11-16 18:26

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('moodle', '0027_alter_historicalmoodleenterprisecustomerconfiguration_options'),
    ]

    operations = [
        migrations.AddField(
            model_name='historicalmoodleenterprisecustomerconfiguration',
            name='enable_incomplete_progress_transmission',
            field=models.BooleanField(default=False, help_text='When set to True, the configured customer will receive learner data transmissions, for incomplete courses as well'),
        ),
        migrations.AddField(
            model_name='moodleenterprisecustomerconfiguration',
            name='enable_incomplete_progress_transmission',
            field=models.BooleanField(default=False, help_text='When set to True, the configured customer will receive learner data transmissions, for incomplete courses as well'),
        ),
    ]
