# Generated by Django 5.2 on 2025-08-28 07:19

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('enterprise', '0240_alter_enterprisecustomer_enable_data_sharing_consent_and_more'),
    ]

    operations = [
        migrations.AlterField(
            model_name='defaultenterpriseenrollmentintention',
            name='realized_enrollments',
            field=models.ManyToManyField(through='enterprise.DefaultEnterpriseEnrollmentRealization', through_fields=('intended_enrollment', 'realized_enrollment'), to='enterprise.enterprisecourseenrollment'),
        ),
    ]
