# Generated by Django 4.2.15 on 2024-09-24 21:39

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('enterprise', '0220_alter_updateroleassignmentswithcustomersconfig_role'),
    ]

    operations = [
        migrations.AddField(
            model_name='enterprisegroup',
            name='group_type',
            field=models.CharField(blank=True, choices=[('budget', 'Budget'), ('flex', 'Flex')], default='flex', help_text='The type of enterprise group', max_length=20, null=True, verbose_name='Group Type'),
        ),
        migrations.AddField(
            model_name='historicalenterprisegroup',
            name='group_type',
            field=models.CharField(blank=True, choices=[('budget', 'Budget'), ('flex', 'Flex')], default='flex', help_text='The type of enterprise group', max_length=20, null=True, verbose_name='Group Type'),
        ),
    ]
