# Generated by Django 4.2.16 on 2024-10-24 15:42

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('enterprise', '0226_alter_enterprisegroup_name_and_more'),
    ]

    operations = [
        migrations.AlterField(
            model_name='defaultenterpriseenrollmentintention',
            name='content_type',
            field=models.CharField(blank=True, choices=[('course', 'Course'), ('course_run', 'Course Run')], help_text='The type of content (e.g. a course vs. a course run).', max_length=127, null=True),
        ),
        migrations.AlterField(
            model_name='historicaldefaultenterpriseenrollmentintention',
            name='content_type',
            field=models.CharField(blank=True, choices=[('course', 'Course'), ('course_run', 'Course Run')], help_text='The type of content (e.g. a course vs. a course run).', max_length=127, null=True),
        ),
        migrations.AddConstraint(
            model_name='defaultenterpriseenrollmentintention',
            constraint=models.UniqueConstraint(fields=('enterprise_customer', 'content_key'), name='unique_default_enrollment_intention'),
        ),
    ]
