# Generated by Django 3.2.22 on 2024-01-25 09:25

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('cornerstone', '0031_cornerstoneapirequestlogs'),
    ]

    operations = [
        migrations.DeleteModel(
            name='CornerstoneAPIRequestLogs',
        ),
    ]
