# Generated by Django 3.2.15 on 2022-09-29 13:39

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('enterprise', '0163_auto_20220928_1611'),
    ]

    operations = [
        migrations.AddField(
            model_name='enterprisecatalogquery',
            name='include_exec_ed_2u_courses',
            field=models.BooleanField(default=False, help_text="Specifies whether the catalog is allowed to include exec ed (2U) courses.  This means that, when the content_filter specifies that 'course' content types should be included in the catalog, executive-education-2u course types won't be excluded from the content of the associated catalog."),
        ),
    ]
