# Generated by Django 3.2.19 on 2023-05-17 15:50

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('enterprise', '0171_auto_20230503_1413'),
    ]

    operations = [
        migrations.AddField(
            model_name='enterprisecourseenrollment',
            name='unenrolled',
            field=models.BooleanField(blank=True, db_index=True, default=None, help_text="Specifies whether the enterprise course enrollment's course enrollment object was unenrolled.", null=True),
        ),
        migrations.AddField(
            model_name='historicalenterprisecourseenrollment',
            name='unenrolled',
            field=models.BooleanField(blank=True, db_index=True, default=None, help_text="Specifies whether the enterprise course enrollment's course enrollment object was unenrolled.", null=True),
        ),
    ]
